library(tidyverse) # version 2.0.0 (dplyr 1.1.1, stringr 1.5.0, tidyr 1.3.0)
library(mirt) # version 1.38.1
library(brms) # version 2.19.0


#### Data Import and Cleaning ####

# read in fake data
d <- read.table("fakedata.txt")

# create dichotomized data set
dd <- d %>% mutate_all(function(x) as.numeric(x > 0))

# long format data for use with brms - for polytomous models, brms count models start counting at 1, not 0
dl <- d %>% as.data.frame %>% mutate(id = 1:n()) %>% 
  pivot_longer(-id, names_to = "item", values_to = "resp") %>%
  mutate(item = factor(item), resp = resp + 1, id = factor(id))

# long format dichotomized
ddl <- dl %>% mutate(resp = as.numeric(resp > 1))



#### MSA Helper Functions ####

## optimization approach
msa_o <- function(mv, pars, pfunc, tseq, weights, idx, q = .5){
  # mv: vector of mean parameter estimates
  # pars: matrix with 1 row per MI draw and columns with item parameters in the order expected by pfunc
  # pfunc: function with three input arguments - p: parameter vector, theta: ability level, idx: defined below 
  #        and outputs the response probability for each category except the lowest category for each item
  # tseq: values of theta at which to evaluate metric stability
  # weights: vector of weights to each response category except the lowest category for an item
  # idx: vector indicating which item (numbered 1, 2, 3, etc.) each parameter belongs to
  # q: quantile at which to summarize Delta_mq draws
  
  f <- function(eta, pars, pfunc, weights, targetp, idx){
    p <- pfunc(pars, eta, idx = idx)
    sqrt(sum(weights * (p - targetp)^2) / max(idx))
  }
  
  res <- lapply(tseq, function(t){
    targetp <- pfunc(p = mv, theta = t, idx = idx)
    apply(pars, 1, function(pars){
      optimize(f = f, interval = c(-6, 6), pars = pars, pfunc = pfunc, weights = weights, targetp = targetp, idx = idx)
    })
  })
  
  res_dist <- t(sapply(res, function(x) sapply(x, function(y) y$objective)))
  
  apply(res_dist, 1, quantile, probs = q)
}

# quantile approach

msa_q <- function(brmsobject, draws, pfunc, qtiles, weights, idx, q = .5){
  ## brmsobject - object of class brmsfit from fitting a model with brms
  ## draws - matrix of MCMC iterations x parameters, number of columns should match the vector length expected by pfunc
  # pfunc: function with three input arguments - p: parameter vector, theta: ability level, idx: defined below 
  #        and outputs the response probability for each category except the lowest category for each item
  # qtiles: theta quantiles at which to evaluate metric stability
  # weights: vector of weights to each response category except the lowest category for an item
  # idx: vector indicating which item (numbered 1, 2, 3, etc.) each parameter belongs to
  # q: quantile at which to summarize Delta_mq draws
  
  thetadraws <- as_draws_matrix(brmsobject, variable = "r_id__theta.*", regex = TRUE) # may need to adjust depending how theta values are stores
  qs <- apply(thetadraws, 1, quantile, probs = qtiles)
  bq <- lapply(1:nrow(qs), function(t) sapply(1:ncol(qs), function(i)
    pfunc(p = as.numeric(draws[i, ]), theta = qs[t, i], idx = idx)))
  
  out <- sapply(bq, function(x){
    targetp <- rowMeans(x)
    quantile(apply(x, 2, function(p) sqrt(sum(weights * (p - targetp)^2) / max(idx))), probs = q)
  })
  out
}

## pfuncs for common models - 
#   these functions input item parameter vectors, scalar theta, and (possibly unused) index of parameter to item
#   and output a vector of response probabilities to all but the lowest category

## user-defined pfuncs may be used to conduct MSA on different models - be sure to format other input to msa_o or msa_q appropriately!

p_2pl <- function(p, theta, idx){ # mirt: itemtype = "2PL" and brms: bernoulli logit
  p <- matrix(p, ncol = 2, byrow = TRUE)
  a <- p[, 1]
  d <- p[, 2]
  plogis(a * theta + d)
}

p_grm <- function(p, theta, idx){ # mirt: itemtype = "graded"
  probs <- unlist(lapply(1:max(idx), function(i){
    pp <- p[idx == i]
    -diff(c(plogis(pp[1] * theta + pp[-1]), 0))
  }))
  probs
}

p_grm2 <- function(p, theta, idx){ # brms: cumulative logit
  probs <- unlist(lapply(1:max(idx), function(i){
    pp <- p[idx == i]
    -diff(c(plogis(pp[1] * theta - pp[-1]), 0))
  }))
  probs
}


p_gpcm <- function(p, theta, idx){ # mirt: itemtype = "gpcm"
  probs <- unlist(lapply(1:max(idx), function(i){
    pp <- p[idx == i]
    num <- c(1, exp(1:(length(pp) - 1) * pp[1] * theta + pp[-1]))
    num[-1] / sum(num) # -1 to omit first category
  }))
  probs
}

p_gpcm2 <- function(p, theta, idx){ # brms: acat logit
  probs <- unlist(lapply(1:max(idx), function(i){
    pp <- p[idx == i]
    num <- pp[1] * theta - pp[-1]
    exp(cumsum(num)) / (1 + sum(exp(cumsum(num))))
  }))
  probs
}

# function to calculate recommended item category weights
wts <- function(data){
  ncat <- apply(data, 2, max) + 1
  rep(1 / (ncat - 1), ncat - 1)
} 


#### Fit Models in mirt and brms ####

## mirt with MML

m2pl <- mirt(dd, 1, itemtype = "2PL", SE = TRUE)
mgrm <- mirt(d, 1, itemtype = "graded", SE = TRUE)
mgpcm <- mirt(d, 1, itemtype = "gpcm", SE = TRUE)

## mirt with MBM
prior_2pl <- mirt.model("THETA = 1-5\nPRIOR = (1-5, a1, lnorm, 0, .5), (1-5, d, norm, 0, 1)")
prior_poly <- mirt.model("THETA = 1-5\nPRIOR = (1-5,a1,lnorm,0,.5),(1-5,d1,norm,0,1),(1-5,d2,norm,0,1), (1-5, d3, norm, 0, 1), (1-4, d4, norm, 0, 1)")

m2pl_mb<- mirt(dd, prior_2pl, itemtype = "2PL", SE = TRUE)
mgrm_mb <- mirt(d, prior_poly, itemtype = "graded", SE = TRUE, GenRandomPars = TRUE) # gen random start values needed because same priors are used for all ordered thresholds
mgpcm_mb <- mirt(d, prior_poly, itemtype = "gpcm", SE = TRUE)

## brms 

twopl_formula <- bf(resp ~ 0 + a * theta + d, a ~ 0 + item, d ~ 0 + item, theta ~ 0 + (1 | id), nl = TRUE)

grm_formula <- bf(resp | thres(gr = item) ~ a * theta, a ~ 0 + item, theta ~ 0 + (1 | id), nl = TRUE)

gpcm_formula <- grm_formula

twopl_prior <- prior("lognormal(0, .5)", class = "b", nlpar = "a", lb = 0) + 
               prior("normal(0, 1)", class = "b", nlpar = "d") +
               prior("constant(1)", class = "sd", nlpar = "theta")

grm_prior <- prior("lognormal(0, .5)", class = "b", nlpar = "a", lb = 0) + 
             prior("normal(0, 1)", class = "Intercept") +
             prior("constant(1)", class = "sd", nlpar = "theta")

gpcm_prior <- grm_prior


twopl_brms <- brm(formula = twopl_formula, data = ddl, family = brmsfamily("bernoulli", "logit"), 
                  prior = twopl_prior, chains = 4, cores = 4, iter = 2000)

grm_brms <- brm(grm_formula, data = dl ,family = brmsfamily("cumulative", "logit"), 
                prior = grm_prior, cores = 4, chains = 4, iter = 2000)

gpcm_brms <- brm(gpcm_formula, data = dl, family = brmsfamily("acat", "logit"), 
                 prior = gpcm_prior, cores = 4, chains = 4, iter = 2000)


#### Extract Information From Models ####

## for models estimated with mirt, extract mean vector and variance-covariance matrix
mv_2pl <- extract.mirt(m2pl, "parvec")
vcv_2pl <- extract.mirt(m2pl, "vcov")
mv_grm <- extract.mirt(mgrm, "parvec")
vcv_grm <- extract.mirt(mgrm, "vcov")
mv_gpcm <- extract.mirt(mgpcm, "parvec")
vcv_gpcm <- extract.mirt(mgpcm, "vcov")

mv_2pl_mb <- extract.mirt(m2pl_mb, "parvec")
vcv_2pl_mb <- extract.mirt(m2pl_mb, "vcov")
mv_grm_mb <- extract.mirt(mgrm_mb, "parvec")
vcv_grm_mb <- extract.mirt(mgrm_mb, "vcov")
mv_gpcm_mb <- extract.mirt(mgpcm_mb, "parvec")
vcv_gpcm_mb <- extract.mirt(mgpcm_mb, "vcov")

# for brms models, extract estimates for each iteration and rearrange columns appropriately (a1, d1_1, d1_2, d1_3, d1_4, a2, ...)
twopl_vars <- bind_cols(lapply(names(d), as_draws_matrix, x = twopl_brms, regex = TRUE))

grm_vars <- lapply(names(d), as_draws_matrix, x = grm_brms, regex = TRUE)
grm_vars <- bind_cols(lapply(grm_vars, function(x) cbind(x[, ncol(x)], x[, -ncol(x)]))) # put a parameters in front

gpcm_vars <- lapply(names(d), as_draws_matrix, x = gpcm_brms, regex = TRUE)
gpcm_vars <- bind_cols(lapply(gpcm_vars, function(x) cbind(x[, ncol(x)], x[, -ncol(x)]))) # put a parameters in front



#### Calculate Stability

R <- 1000 # number of MIs

tseq <- seq(-4, 4, by = .25) # for optimization approach
quantiles <- seq(.05, .95, by = .05) # for brms quantile approach

# item indices
idx_2pl <- cumsum(substr(colnames(vcv_2pl), 1, 1) == "a")
idx_grm <- cumsum(substr(colnames(vcv_grm), 1, 1) == "a")
idx_gpcm <- cumsum(substr(colnames(vcv_gpcm), 1, 1) == "a")


msa_2pl <- msa_o(mv = mv_2pl, pars = MASS::mvrnorm(R, mv_2pl, vcv_2pl),
                 pfunc = p_2pl, tseq = tseq, weights = wts(dd), idx = idx_2pl)

msa_grm <- msa_o(mv = mv_grm, pars = MASS::mvrnorm(R, mv_grm, vcv_grm),
                 pfunc = p_grm, tseq = tseq, weights = wts(d), idx = idx_grm)

msa_gpcm <- msa_o(mv = mv_gpcm, pars = MASS::mvrnorm(R, mv_gpcm, vcv_gpcm),
                  pfunc = p_gpcm, tseq = tseq, weights = wts(d), idx = idx_gpcm)

msa_2pl_mb <- msa_o(mv = mv_2pl, pars = MASS::mvrnorm(R, mv_2pl_mb, vcv_2pl_mb),
                    pfunc = p_2pl, tseq = tseq, weights = wts(dd), idx = idx_2pl)

msa_grm_mb <- msa_o(mv = mv_grm_mb, pars = MASS::mvrnorm(R, mv_grm_mb, vcv_grm_mb),
                    pfunc = p_grm, tseq = tseq, weights = wts(d), idx = idx_grm)

msa_gpcm_mb <- msa_o(mv = mv_gpcm_mb, pars = MASS::mvrnorm(R, mv_gpcm_mb, vcv_gpcm_mb),
                     pfunc = p_gpcm, tseq = tseq, weights = wts(d), idx = idx_gpcm)

msa_2pl_brmsO <- msa_o(mv = colMeans(twopl_vars), pars = twopl_vars, pfunc = p_2pl, 
                       tseq = tseq, weights = wts(dd), idx = idx_2pl)

msa_grm_brmsO <- msa_o(mv = colMeans(grm_vars), pars = grm_vars, pfunc = p_grm2,
                       tseq = tseq, weights = wts(d), idx = idx_grm)

msa_gpcm_brmsO <- msa_o(mv = colMeans(gpcm_vars), pars = gpcm_vars, pfunc = p_gpcm2,
                        tseq = tseq, weights = wts(d), idx = idx_gpcm)


msa_2pl_brmsQ <- msa_q(brmsobject = twopl_brms, draws = twopl_vars, 
                       pfunc = p_2pl, qtiles = quantiles, weights = wts(dd), idx = idx_2pl)
msa_grm_brmsQ <- msa_q(brmsobject = grm_brms, draws = grm_vars, 
                       pfunc = p_grm2, qtiles = quantiles, wts(d), idx = idx_grm)
msa_gpcm_brmsQ <- msa_q(brmsobject = gpcm_brms, draws = gpcm_vars, 
                        pfunc = p_gpcm2,  qtiles = quantiles, wts(d), idx = idx_gpcm)


#### Visualize Results ####

# all 2PL 
plot(tseq, msa_2pl, ylim = c(0, .06), type = 'l')
lines(tseq, msa_2pl_mb)
lines(tseq, msa_2pl_brmsO)
lines(qnorm(quantiles), msa_2pl_brmsQ)


# all GRM
plot(tseq, msa_grm, ylim = c(0, .1), type = 'l')
lines(tseq, msa_grm_mb)
lines(tseq, msa_grm_brmsO)
lines(qnorm(quantiles), msa_grm_brmsQ)

# all GPCM
plot(tseq, msa_gpcm, ylim = c(0, .1), type = 'l')
lines(tseq, msa_gpcm_mb)
lines(tseq, msa_gpcm_brmsO)
lines(qnorm(quantiles), msa_gpcm_brmsQ)
